"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.isJson = void 0;
const fs_1 = require("fs");
const path_1 = require("path");
const rr = require('railroad-diagrams');
const m = require('mustache');
let dodebug = false;
function railroad(grammar, out, opts) {
    let settings = opts;
    if (opts.config && isJson(fs_1.readFileSync(path_1.resolve(process.cwd(), opts.config), "utf8"))) {
        settings = JSON.parse(fs_1.readFileSync(path_1.resolve(process.cwd(), opts.config), "utf8"));
    }
    settings = {
        template: settings.template || path_1.resolve(__dirname, "../../template")
    };
    if (!fs_1.existsSync(path_1.resolve(process.cwd(), grammar))) {
        console.log(`${path_1.resolve(process.cwd(), grammar)} does not exist. Please make sure you are passing the correct path to your grammar file.`);
        process.exit(1);
    }
    if (!fs_1.existsSync(path_1.resolve(settings.template, "index.html"))) {
        console.log(`${path_1.resolve(settings.template, "index.html")} does not exist. Please make sure you are passing the correct path to your grammar file.`);
        process.exit(1);
    }
    let template = fs_1.readFileSync(path_1.resolve(settings.template, "index.html"), 'utf8');
    grammar = require(path_1.resolve(process.cwd(), grammar));
    let rules = new Map();
    if (opts.debug) {
        dodebug = true;
    }
    for (let rule of grammar.ParserRules) {
        if (rules.has(rule.name) && rules.get(rule.name) !== undefined) {
            // @ts-ignore
            let oldrules = rules.get(rule.name);
            oldrules.push(rule);
            rules.set(rule.name, oldrules);
        }
        else {
            rules.set(rule.name, [rule]);
        }
    }
    let diagrams = [];
    diagrams.push(...traceandgram(rules, grammar.ParserStart, [], false));
    diagrams.reverse();
    diagrams[0].comment = "- Start";
    let rulekeys = [];
    for (let rule of rules.keys()) {
        rulekeys.push(rule);
    }
    let page = m.render(template, {
        name: "Test",
        start: grammar.ParserStart,
        diagrams,
        ruleslist: JSON.stringify(rulekeys)
    });
    fs_1.writeFileSync(path_1.resolve(process.cwd(), out), page);
    console.log("Done!");
}
exports.default = railroad;
function traceandgram(rules, start, diagrammed, issub, sub) {
    let toreturn = [];
    let rule;
    if (rules.get(start) === undefined) {
        rule = [];
    }
    else {
        // @ts-ignore
        rule = rules.get(start);
    }
    if (sub) {
        rule = [sub];
    }
    let bits = [];
    if (rule !== undefined) {
        if (dodebug)
            console.log(`Scanning rule "${start}"...`);
        if (rule.length === 1) {
            let bit = rule[0];
            for (let symbol of bit.symbols) {
                if (typeof symbol === "string" && /[A-z]\$ebnf\$[0-9][0-9]?/.test(symbol)) {
                    let ebnf = rules.get(symbol);
                    if (ebnf !== undefined) {
                        if (ebnf[0].symbols.length > 0 && ebnf[1].symbols.length === 0) {
                            if (dodebug)
                                console.log(`Rule "${symbol}" is an optional ebnf rule`);
                            ebnf[0].symbols.forEach((symboll) => {
                                if (typeof symboll === "string" && !diagrammed.includes(symboll) && rules.has(symboll)) {
                                    if (dodebug)
                                        console.log(`Came across new rule "${symboll}" in rule "${symbol}"`);
                                    toreturn.push(...traceandgram(rules, symboll, diagrammed, false));
                                }
                            });
                            bits.push(rr.Optional(rr.Sequence(...traceandgram(rules, ebnf[0].name, diagrammed, true, ebnf[0]))));
                        }
                        else if (ebnf[0].symbols[0] === ebnf[1].symbols[1] && ebnf[1].symbols[0] === ebnf[0].name) {
                            if (dodebug)
                                console.log(`Rule "${symbol}" is a loop ebnf rule`);
                            ebnf[0].symbols.forEach((symboll) => {
                                if (typeof symboll === "string" && !diagrammed.includes(symboll) && rules.has(symboll)) {
                                    if (dodebug)
                                        console.log(`Came across new rule "${symboll}" in rule "${symbol}"`);
                                    toreturn.push(...traceandgram(rules, symboll, diagrammed, false));
                                }
                            });
                            bits.push(rr.OneOrMore(rr.Sequence(...traceandgram(rules, ebnf[0].name, diagrammed, true, ebnf[0]))));
                        }
                        else if (ebnf[0].symbols.length === 0 && ebnf[1].symbols[0] === ebnf[0].name && ebnf[1].symbols.length > 1) {
                            if (dodebug)
                                console.log(`Rule "${symbol}" is a loop and optional ebnf rule`);
                            let copy = ebnf[1];
                            copy.symbols.shift();
                            copy.symbols.forEach((symboll) => {
                                if (typeof symboll === "string" && !diagrammed.includes(symboll) && rules.has(symboll)) {
                                    if (dodebug)
                                        console.log(`Came across new rule "${symboll}" in rule "${symbol}"`);
                                    toreturn.push(...traceandgram(rules, symboll, diagrammed, false));
                                }
                            });
                            bits.push(rr.Optional(rr.OneOrMore(rr.Sequence(...traceandgram(rules, ebnf[0].name, diagrammed, true, copy)))));
                        }
                    }
                }
                else if (typeof symbol === "string" && /[A-z]\$subexpression\$[0-9][0-9]?/.test(symbol)) {
                    let ebnf = rules.get(symbol);
                    if (ebnf !== undefined) {
                        if (dodebug)
                            console.log(`Rule "${symbol}" is a subexpression`);
                        if (ebnf[0].symbols.length > 0 && ebnf[1].symbols.length > 0) {
                            let subchoices = [];
                            ebnf.forEach((value) => {
                                subchoices.push(rr.Choice(0, ...traceandgram(rules, "", diagrammed, true, value)));
                            });
                            bits.push(rr.Sequence(...subchoices));
                        }
                    }
                }
                else if (typeof symbol === "string") {
                    if (dodebug)
                        console.log(`Rule "${symbol}" is a non-terminal string`);
                    bits.push(rr.NonTerminal(symbol));
                    if (!diagrammed.includes(symbol) && rules.has(symbol)) {
                        diagrammed.push(symbol);
                        toreturn.push(...traceandgram(rules, symbol, diagrammed, false));
                    }
                }
                else if (symbol.type) {
                    if (dodebug)
                        console.log(`Rule "${symbol.type}" is a non-terminal lexer value`);
                    bits.push(rr.NonTerminal("%" + symbol.type));
                }
                else if (symbol.literal) {
                    if (dodebug)
                        console.log(`Rule "${symbol.literal}" is terminal/literal`);
                    bits.push(rr.Terminal(symbol.literal));
                }
            }
        }
        else if (rule.length > 1) {
            let bitss = [];
            rule.forEach((bit) => {
                let topush = [];
                bit.symbols.map((symbol) => {
                    if (typeof symbol === "string" && /[A-z]\$ebnf\$[0-9][0-9]?/.test(symbol)) {
                        let ebnf = rules.get(symbol);
                        if (ebnf !== undefined) {
                            if (ebnf[0].symbols.length > 0 && ebnf[1].symbols.length === 0) {
                                if (dodebug)
                                    console.log(`Rule "${symbol}" is an optional ebnf rule`);
                                ebnf[0].symbols.forEach((symboll) => {
                                    if (typeof symboll === "string" && !diagrammed.includes(symboll) && rules.has(symboll)) {
                                        if (dodebug)
                                            console.log(`Came across new rule "${symboll}" in rule "${symbol}"`);
                                        toreturn.push(...traceandgram(rules, symboll, diagrammed, false));
                                    }
                                });
                                topush.push(rr.Optional(rr.Sequence(...traceandgram(rules, ebnf[0].name, diagrammed, true, ebnf[0]))));
                            }
                            else if (ebnf[0].symbols[0] === ebnf[1].symbols[1] && ebnf[1].symbols[0] === ebnf[0].name) {
                                if (dodebug)
                                    console.log(`Rule "${symbol}" is a loop ebnf rule`);
                                ebnf[0].symbols.forEach((symboll) => {
                                    if (typeof symboll === "string" && !diagrammed.includes(symboll) && rules.has(symboll)) {
                                        if (dodebug)
                                            console.log(`Came across new rule "${symboll}" in rule "${symbol}"`);
                                        toreturn.push(...traceandgram(rules, symboll, diagrammed, false));
                                    }
                                });
                                topush.push(rr.OneOrMore(rr.Sequence(...traceandgram(rules, ebnf[0].name, diagrammed, true, ebnf[0]))));
                            }
                            else if (ebnf[0].symbols.length === 0 && ebnf[1].symbols[0] === ebnf[0].name && ebnf[1].symbols.length > 1) {
                                if (dodebug)
                                    console.log(`Rule "${symbol}" is a loop and optional ebnf rule`);
                                let copy = ebnf[1];
                                copy.symbols.shift();
                                copy.symbols.forEach((symboll) => {
                                    if (typeof symboll === "string" && !diagrammed.includes(symboll) && rules.has(symboll)) {
                                        if (dodebug)
                                            console.log(`Came across new rule "${symboll}" in rule "${symbol}"`);
                                        toreturn.push(...traceandgram(rules, symboll, diagrammed, false));
                                    }
                                });
                                bits.push(rr.Optional(rr.OneOrMore(rr.Sequence(...traceandgram(rules, ebnf[0].name, diagrammed, true, copy)))));
                            }
                        }
                    }
                    else if (typeof symbol === "string" && /[A-z]\$subexpression\$[0-9][0-9]?/.test(symbol)) {
                        let ebnf = rules.get(symbol);
                        if (ebnf !== undefined) {
                            if (dodebug)
                                console.log(`Rule "${symbol}" is a subexpression`);
                            if (ebnf[0].symbols.length > 0 && ebnf[1].symbols.length > 0) {
                                let subchoices = [];
                                ebnf.forEach((value) => {
                                    subchoices.push(rr.Choice(0, ...traceandgram(rules, "", diagrammed, true, value)));
                                });
                                topush.push(rr.Sequence(...subchoices));
                            }
                        }
                    }
                    else if (typeof symbol === "string") {
                        if (dodebug)
                            console.log(`Rule "${symbol}" is a non-terminal string`);
                        topush.push(rr.NonTerminal(symbol));
                        if (!diagrammed.includes(symbol) && rules.has(symbol)) {
                            diagrammed.push(symbol);
                            toreturn.push(...traceandgram(rules, symbol, diagrammed, false));
                        }
                    }
                    else if (symbol.type) {
                        if (dodebug)
                            console.log(`Rule "${symbol.type}" is a non-terminal lexer value`);
                        topush.push(rr.NonTerminal("%" + symbol.type));
                    }
                    else if (symbol.literal) {
                        if (dodebug)
                            console.log(`Rule "${symbol.literal}" is terminal/literal`);
                        topush.push(rr.Terminal(symbol.literal));
                    }
                });
                bitss.push(rr.Sequence(...topush));
            });
            bits.push(rr.Choice(0, ...bitss));
        }
    }
    if (issub) {
        return bits;
    }
    else {
        toreturn.push({
            name: start,
            diagram: rr.Diagram(...bits).toString(),
        });
        return toreturn;
    }
}
function isJson(data) {
    try {
        JSON.parse(data);
    }
    catch (e) {
        return false;
    }
    return true;
}
exports.isJson = isJson;
//# sourceMappingURL=railroad.js.map